module Kristall_Modul

! Allgemeine Parameter
integer,parameter::DIM=3                       ! Raumdimension (2D oder 3D)
integer,parameter::NACHBARN=3**DIM
logical,parameter::ABSCHNEIDEN=.true.

integer,parameter::AUSGEBEN_ALLE=20
real,parameter::   DELTA_T=0.01       ! Zeitschrittweite
real,parameter::   T_MAX=55.0
          ! Endzeit
! Physikalische Konstanten
real,parameter:: EPS=5.0
real,parameter:: SIGMA=1.0
real,parameter:: R_AB=2.5*SIGMA
real,parameter:: MASSE=1.0
real,parameter:: TEMP= 0.5
integer,parameter:: NX= 6
integer,parameter:: NY= 6
integer,parameter:: NZ= 6
real,parameter:: RGG=2**(1.0/6.0)*SIGMA
integer,parameter::NGX=1.4*NX
integer,parameter::NGY=1.4*NY
integer,parameter:: NGZ= 1.4*NZ
! Eigene Datentypen
type Atom
 real::m                                !   Masse
  real,dimension(DIM)::x,v,F,F_alt      !   Ort,Geschwindigkeit,Kraft
  integer::zelle
  integer::typ                          !   Typ (fuer Darstellung)
end type
type Atome                               ! Sonnensystem:
  integer::N                             !   Anzahl der Himmelskoerper
  type(Atom),dimension(:),pointer::a     !   Feld der Himmelskoerper
end type

type Gebiet
integer::N,Nx,Ny,Nz
real:: x0,y0,z0
real::d
integer,dimension(:,:),pointer::Nachbar
integer,dimension(:),pointer::Anzahl,Start
end type

contains

function Zellnummer(G,A,i)
      integer::Zellnummer; type(Gebiet)::G; type(Atome)::A; integer::i
  integer::zx,zy,zz
  zx=int((A%a(i)%x(1)-G%x0)/G%d)
  zy=int((A%a(i)%x(2)-G%y0)/G%d)
  zz=int((A%a(i)%x(3)-G%z0)/G%d)
  if ( (0<=zx) .and. (zx<G%Nx) .and. (0<=zy) .and. (zy<G%Ny) .and. (0<=zz) .and. (zz<G%Ny)) then
    Zellnummer=1+zx+zy*G%Nx+zz*G%Nx*G%Ny
  else
    Zellnummer=0
  end if
end function

function Zellen_benachbart(G,i,j)
      logical::Zellen_benachbart; type(Gebiet)::G; integer::i,j
  integer::ix,iy,iz,jx,jy,jz,k

  k=i-1; ix=mod(k,G%Nx); k=k/G%Nx; iy=mod(k,G%Ny); iz=k/G%Ny
  k=j-1; jx=mod(k,G%Nx); k=k/G%Nx; jy=mod(k,G%Ny); jz=k/G%Ny

  if ( (i/=0) .and. (j/=0) .and. (abs(ix-jx)<=1) .and. (abs(iy-jy)<=1) .and. (abs(iz-jz)<=1)) then
    Zellen_benachbart=.true.
  else
    Zellen_benachbart=.false.
  end if
end function

subroutine Atome_sortieren(G,A)
      type(Gebiet)::G; type(Atome)::A
  integer::i,k,l,z,zn
  type(Atom)::aa 
  ! Anzahl und Startposition fuer jedes Zelle bestimmen
  G%Anzahl(:)=0
  do k=1,A%N
    i=A%a(k)%zelle 
    G%Anzahl(i)=G%Anzahl(i)+1
  end do
  G%Start(0)=1
  do i=1,G%N
    G%Start(i)=G%Start(i-1)+G%Anzahl(i-1)
  end do
  ! Atome einsortieren
  G%Anzahl(:)=0    ! Zahl der richtig einsortierten Atome
  z=0              ! Aktuelle Zelle
  do while (z<G%N)      ! "<", da letzte Zelle automatisch richtig
    ! Solange noch Teilchen in Zelle z fehlen, durch Tauschen welche suchen
    do while ( G%Start(z)+G%Anzahl(z)<G%Start(z+1) )
      k=G%Start(z)+G%Anzahl(z)   ! erstes potentiell falsche Atom in Zelle z
      zn=A%a(k)%zelle            ! Zelle, wo das aktuelle Atom hin muss
      do while (zn/=z)
        ! aktuelles Atom richtig einsortieren und dortiges Atom herholen
        l=G%Start(zn)+G%Anzahl(zn)   ! hier muss das aktuelle Atom hin
        G%Anzahl(zn)=G%Anzahl(zn)+1  ! und hier ist es dann auch richtig
        aa=A%a(l); A%a(l)=A%a(k); A%a(k)=aa  ! Atome k und l tauschen
        zn=A%a(k)%zelle              ! Gebiet des geholten Atoms
      end do
      G%Anzahl(z)=G%Anzahl(z)+1  ! aktuelles Teilchen ist richtig einsortiert
    end do
    z=z+1               ! aktuelle Zelle fertig, also zur naechsten
  enddo
  ! letze Zelle wurde nicht ganz durchlaufen, deshalb noch Anzahl berechnen
  G%Anzahl(G%N)=A%N-G%Start(G%N)+1
end subroutine

subroutine Zellen_initialisieren(G)
      type(Gebiet)::G
  integer::i,j,k
  ! Zelleninitialisieren, Zelle 0 fr verlorene Teilchen
  G%x0=-0.5*NGX
  G%y0=-0.5*NGY
  G%z0=-0.5*NGZ
  G%d =R_AB
  G%Nx=int(NGX*RGG/G%d)+1
  G%Ny=int(NGY*RGG/G%d)+1
  G%Nz=int(NGZ*RGG/G%d)+1
  G%N=G%Nx*G%Ny*G%Nz                         ! Zahl der echten Zellen
  allocate (G%Anzahl(0:G%N))
  allocate (G%Start(0:G%N))
  allocate (G%Nachbar(0:G%N,NACHBARN))
  G%Nachbar(:,:)=-1       ! evtl. nicht vorhandene Nachbarn durch -1 markieren
  do i=0,G%N
    k=0                   ! Zahl der Nachbarzellen von Zelle i
    do j=0,G%N
      if(Zellen_benachbart(G,i,j)) then
        k=k+1
        G%Nachbar(i,k)=j  ! k-ter Nachbar von Zelle i ist j
      end if
    end do
  end do
end subroutine

! ***  Zeitintegration und Ausgabe  *******************************************
function NV_Zufallszahl()
    real::NV_Zufallszahl
    real:: a1,a2,r
    r=0
    do while ((r==0).or.(1<=r))
	call random_number(a1); a1=2*a1-1
	call random_number(a2); a2=2*a2-1
	r=a1**2+a2**2
    end do
    NV_Zufallszahl=a1*sqrt(-2*log(r)/r)
end function


subroutine Temperatur (A,T)
 type(Atome)::A
 real::T
 integer::i,d
 real::s,z
  do i=1,A%N
    s=sqrt(T/A%a(i)%m)
     do d=1,DIM
       z=NV_Zufallszahl()*s
       A%a(i)%v(d)=A%a(i)%v(d)+z
     end do
  end do
end subroutine
function Temperatur_Ausgabe(A)
 type(Atome)::A
 real::Temperatur_Ausgabe
 integer::i
 real::s
  s=0
  do i=1,A%N
    s=s+0.5*A%a(i)%m*sum(A%a(i)%v**2)
  end do
 Temperatur_Ausgabe=s
end function
subroutine Temperatur_anpassen(A,T)
  type(Atome)::A
      real::T,Ta,F
  integer::i
  Ta=Temperatur_Ausgabe(A)
  F=sqrt(T/Ta)
  do i=1,A%N
    A%a(i)%v=A%a(i)%v*F
  end do
end subroutine

subroutine Ausgabe(A,t)
      type(Atome)::A
      real::t
  integer::i
  write(unit=1,fmt=*) A%N
  write(unit=1,fmt=*) t," fs"
  do i=1,A%N
    write(unit=1,fmt=*) A%a(i)%typ, A%a(i)%x, 0
  end do
end subroutine

subroutine Kraft(A,B)
      type(Atom)::A,B
  real,dimension(DIM)::xij
  real::rr,zsdr6
  xij=B%x-A%x                  ! Abstandsvektor von i nach j
  rr=sum(xij**2)                                 ! r^2
  zsdr6=2*(SIGMA**2/rr)**3
  xij=12*EPS/rr*zsdr6*(1-zsdr6)*xij                 ! Kraft von j auf i
  A%F=A%F+xij
  B%F=B%F-xij
end subroutine
subroutine Berechne_F(G,A)
      type(Gebiet)::G
      type(Atome)::A
integer::j,k,i,zn,z
  do k=1,A%N
    A%a(k)%F_alt=A%a(k)%F
    A%a(k)%F=0
  end do
  do k=1,A%N
    z= A%a(k)%zelle
    do i=1,NACHBARN
      zn=G%Nachbar(z,i)                  ! zn ist der i-te NAchbar von z
      if(zn/=-1)then
        do j=G%Start(zn),G%Start(zn)+G%Anzahl(zn)-1
          if (j<k) call Kraft(A%a(k),A%a(j))
        end do
      end if
    end do
  end do
end subroutine

subroutine Berechne_x(A,dt)
      type(Atome)::A; real::dt
  integer::i
  do i=1,A%N
    A%a(i)%x=A%a(i)%x+dt*A%a(i)%v+dt**2/(2*A%a(i)%m)*A%a(i)%F
  end do
end subroutine

subroutine Berechne_v(A,dt)
      type(Atome)::A; real::dt
  integer::k
  do k=1,A%N
    A%a(k)%v=A%a(k)%v+dt*(A%a(k)%F+A%a(k)%F_alt)/(2*A%a(k)%m)
  end do
end subroutine

subroutine Zeitintegration(G,A,dt,t_end)
type(Gebiet)::G
type(Atome)::A; real::dt,t_end
  integer::k,j
  real::t
  t=0
  k=0
  do j=1,A%N
    A%a(j)%zelle=Zellnummer(G,A,j)
  end do
  call Atome_sortieren(G,A)                                 ! Zaehler fuer Ausgabe
  call Berechne_F(G,A)
  do while(t<t_end)
    t=t+dt
    call Berechne_x(A,dt)
    do j=1,A%N
      A%a(j)%zelle=Zellnummer(G,A,j)
    end do
    call Atome_sortieren(G,A)
    call Berechne_F(G,A)
    call Berechne_v(A,dt)
    call Temperatur_anpassen(A,TEMP)
    k=k+1
    if(mod(k,AUSGEBEN_ALLE)==0) call Ausgabe(A,t)
  end do
end subroutine

! ***  Startkonfiguration  ****************************************************

subroutine Startkonfiguration(A)
      type(Atome)::A
 integer::u,v,k,w
! Partikel initialisieren
  A%N=(NX+1)*(NY+1)*(NZ+1)
  allocate(A%a(A%N))
  k=0
  Do u=0,NX
     do v=0,NY
       do w=0,NZ
         k=k+1
         A%a(k)%x(1)=(u-NX/2)*RGG
         A%a(k)%x(2)=(v-NY/2)*RGG
         A%a(k)%x(3)=(w-NZ/2)*RGG
         A%a(k)%m=MASSE
         A%a(k)%v=0
       end do
     end do
  end do
end subroutine

end module

! ***  Hauptprogramm  *********************************************************

program Kristall
use Kristall_Modul

type(Atome)::A
type(Gebiet)::G

open(unit=1,file="kristall_3D.xyz",action="write",status="replace")
call Zellen_initialisieren(G)
call Startkonfiguration(A)
call Temperatur(A,TEMP)
call Ausgabe(A,0.0)

call Zeitintegration(G,A,DELTA_T,T_MAX)

close(unit=1)

deallocate(A%a)
deallocate(G%Anzahl)
deallocate(G%Start)
deallocate(G%Nachbar)

end program 
